{-|
Module      : Data.Graph.Clustering.FLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

# Louvain Algorithm

Our algorithm is divided into two phases that are repeated iteratively.
Assume that we start with a weighted network of N nodes. First, we
assign a different community to each node of the network. So, in this
initial partition there are as many communities as there are nodes.
Then, for each node i we consider the neighbours j of i and we evaluate
the gain of modularity that would take place by removing i from its
community and by placing it in the community of j. The node i is then
placed in the community for which this gain is maximum (in the case of
a tie we use a breaking rule), but only if this gain is positive. If
no positive gain is possible, i stays in its original community. This
process is applied repeatedly and sequentially for all nodes until
no further improvement can be achieved and the first phase is then
complete.

Reference:
Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud;
Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities
in large networks". Journal of Statistical Mechanics: Theory and
Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible.
doi:10.1088/1742-5468/2008/10/P10008.

-}

module Data.Graph.Clustering.FLouvain
  where

import Protolude
import Data.Graph.Inductive
import qualified Data.List as DL

import Data.Graph.FGL

--  "glue" : function to gather/merge communities
--  "klue" : function to split communities
data ClusteringMethod = Glue | Klue
  deriving (Eq)

------------------------------------------------------------------------
-- | Fixed point with at most n iterations
-- 'Int' argument is the maximal number of iterations to make
-- 'a -> a' is the iterator function
-- 'a -> Bool' is the condition checking function ('True' continues looping, 'False' breaks it)
-- 'a' is the initial value
fixPt :: Int -> (a -> a) -> (a -> Bool) -> a -> a
fixPt 0 iterator _ init = iterator init
fixPt n iterator cond init = 
  if cond next
     then fixPt (n - 1) iterator cond init
     else next
  where
    next = iterator init

-- | Main Louvain first step iteration function
louvainFirstStepIterate :: Int -> FGraph a b -> CGr
louvainFirstStepIterate n gr = fixPt n iterator cond initCGr
  where
    initCGr      = initialCGr  gr
    grWeight     = graphWeight gr
    iterator cgr = iteration   gr cgr
    cond cgr     = (unModularity $ modularity gr cgr grWeight) < 0.1

------------------------------------------------------------------------

-- Our basic graph. Nodes have custom labels. Edges have weight assigned to them.
newtype Weight = Weight { unWeight :: Double }
  deriving (Show, Eq, Ord)
type FEdge b = (Weight, b)
fedgeWeight :: FEdge b -> Double
fedgeWeight = unWeight . fst
sumEdgeWeights :: Adj (FEdge b) -> Double
sumEdgeWeights es = sum $ map (fedgeWeight . fst) es
type FGraph a b = Gr a (FEdge b)

-- | Used for k_i in formula (2)
-- (sum of the weights of the links incident to node i)
newtype NodeWeightSum = NodeWeightSum { unNodeWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeWeightSum :: Context a (FEdge b) -> NodeWeightSum
nodeWeightSum (p, _, _, s) = NodeWeightSum $ sumEdgeWeights $ p <> s

-- |Used for k_i,in in formula (2)
-- (Sum of weights of links from a given 'Node' to nodes in a given 'Community')
newtype NodeComWeightSum = NodeComWeightSum { unNodeComWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeComWeightSum :: Community -> Context a (FEdge b) -> NodeComWeightSum
nodeComWeightSum com (p, _, _, s) =
  NodeComWeightSum $ sumEdgeWeights $ filter (\(_, n) -> n `elem` comNodes com) $ p <> s

newtype NodeNonComWeightSum = NodeNonComWeightSum { unNodeNonComWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeNonComWeightSum :: Community -> Context a (FEdge b) -> NodeNonComWeightSum
nodeNonComWeightSum com (p, _, _, s) =
  NodeNonComWeightSum $ sumEdgeWeights $ filter (\(_, n) -> n `notElem` comNodes com) $ p <> s

-- Probably this structure is better to reduce the number of computations
-- (precompute sum of node weights, which is the k_i variable in formula (2)).
-- type FNode a = (NodeWeightSum, a)
-- fnodeWeightSum :: FNode a -> Double
-- fnodeWeightSum = unNodeWeightSum . fst

-- | This is the m variable in formula (2) of the Louvain paper
newtype GraphWeightSum = GraphWeightSum { unGraphWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_in in formula (2) of the Louvain paper
-- (sum of the weights of the links inside C)
newtype InWeightSum = InWeightSum { unInWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_tot in formula (2) of the Louvain paper
-- (sum of the weights of the links incident to nodes in C)
newtype TotWeightSum = TotWeightSum { unTotWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | Computed Delta_Q value in (2)
newtype DeltaQ = DeltaQ { unDeltaQ :: Double }
  deriving (Show, Eq, Ord)
newtype Modularity = Modularity { unModularity :: Double }
  deriving (Show, Eq, Ord)

-- | Type for the clusters we will be creating.
newtype Community = Community { unCommunity ::  ([Node], InWeightSum, TotWeightSum) }
  deriving (Show, Eq, Ord)
comNodes :: Community -> [Node]
comNodes (Community (ns, _, _)) = ns
comInWeightSum :: Community -> InWeightSum
comInWeightSum (Community (_, inWeightSum, _)) = inWeightSum
comTotWeightSum :: Community -> TotWeightSum
comTotWeightSum (Community (_, _, totWeightSum)) = totWeightSum

type CGrNode = Node
type CGrEdge = (InWeightSum, TotWeightSum)

type CGr = Gr Community ()

graphWeight :: FGraph a b -> GraphWeightSum
graphWeight gr = GraphWeightSum $ 0.5 * ufold (\(_, n, _, _) -> weight' $ context gr n) 0 gr
  where
    weight' (p, _, _, s) acc = acc + (sumEdgeWeights $ p <> s)

-- | Compute initial 'CGr' for a given 'FGraph a b'. This means, put each node
-- in a separate community.
initialCGr :: FGraph a b -> CGr
initialCGr gr = gmap singletonCom gr
  where
    -- A singleton community is given:
    -- the same node id for a community
    -- same incoming/outgoing edges
    -- custom Community label
    singletonCom (p, v, _, s) = (p', v, Community ([v], iws, tws), s')
      where
        p' = map edgeComRemap p
        s' = map edgeComRemap s
        edgeComRemap (_, n) = ((), n)
        edges = lneighbors gr v
        -- no internal links
        iws = InWeightSum 0.0
        -- Just sum over the edges coming into/out of node v.  This is because
        -- there are no inner links, all other nodes are external.
        tws = TotWeightSum $ sumEdgeWeights edges

-- ALGORITHM

-- | Q function from Louvain paper (1).
-- We just fold over the communities (this is because of the delta(c_i, c_j)
-- param)
modularity :: FGraph a b -> CGr -> GraphWeightSum -> Modularity
modularity gr cgr m = Modularity $ coeff * ( ufold modularity' 0.0 cgr )
  where
    coeff = 0.5 / (unGraphWeightSum m)
    -- Sum over nodes in community
    -- \Sum A_ij is just the InWeightSum (in formula (1) it is sum of weights
    -- between nodes i and j both in the same community, as enforced by the
    -- \delta(c_i, c_j) symbol)
    modularity' (_, _, com, _) acc = acc + component
      where
        component = (unInWeightSum $ comInWeightSum com) - weightsMul
        weightsMul = coeff * ( sum $ map weightsMul' $ comNodes com )
        weightsMul' n = (ki n) * (sum $ map ki $ comNodes com)
        -- k_i variable in formula (1)
        ki :: Node -> Double
        ki n = unNodeWeightSum $ nodeWeightSum $ context gr n

type Delta = Community -> NodeWeightSum -> NodeComWeightSum -> GraphWeightSum -> DeltaQ
-- | Delta Q function from Louvain paper (2).
delta :: Delta
delta com ki kiin m = DeltaQ $ acc - dec
  where
    inWeightSum = comInWeightSum com
    totWeightSum = comTotWeightSum com
    acc = accL - accR * accR
    accL = 0.5 * (unInWeightSum inWeightSum + 2.0 * (unNodeComWeightSum kiin)) / (unGraphWeightSum m)
    accR = 0.5 * (unTotWeightSum totWeightSum + unNodeWeightSum ki) / (unGraphWeightSum m)
    dec = decL - decM * decM - decR * decR
    decL = 0.5 * (unInWeightSum inWeightSum) / (unGraphWeightSum m)
    decM = 0.5 * (unTotWeightSum totWeightSum) / (unGraphWeightSum m)
    decR = 0.5 * (unNodeWeightSum ki) / (unGraphWeightSum m)

-- | One iteration step takes the graph and existing communities as a graph and
-- computes new community graph

-- NOTE: xdfsFoldWith only iterates with the defined Context. In each step of
-- the algorithm, we need full node information however (i.e. all edges
-- connected to a node), so instead of calling just 'step' we are forced to call via:
-- 'step . context gr . node''
-- We could avoid the higher complexity, eg. by precomputing the whole graph
-- into a HashMap Node [Edge].
iteration :: FGraph a b -> CGr -> CGr
iteration gr cs = xdfsFoldWith suc' (\(_, v, _, _)
                -> step gw $ context gr $ v) cs (nodes gr) gr
  where
    gw = graphWeight gr
    --weightSum = ufold weightSum' 0 gr
    --weightSum' (p, v, l, s) acc = acc + (sum )

-- TODO Remember to filter out empty Communities
-- | Step for one node. We try re-assign it to a neighbouring community, where
-- the increase of modularity for graph will be the largest
step :: GraphWeightSum -> CFunFold a (FEdge b) CGr
step gw ctx@(_, v, _, _) cgr = newCgr
  where
    newCgr = case mNc of
      Nothing -> cgr
      Just nc ->
        if bestFitdq > 0.0 then
          let newBestFitCom = makeMove Into bestFitCom
              newNc = makeMove OutOf nc
              in
            replaceLNode (replaceLNode cgr newNc) newBestFitCom
        else
          cgr

    (bestFitCom, DeltaQ bestFitdq) =
      maximumBy (\(_, deltaq1) (_, deltaq2) -> compare deltaq1 deltaq2) deltas

    mNc :: Maybe (LNode Community)
    mNc = nodeCommunity v cgr
    ncs :: [LNode Community]
    ncs = nodeNeighbours v cgr

    -- We move node from community nc into ncs
    -- moves :: Maybe (LNode Community, [LNode Community])
    -- moves = case mNc of
    --   Nothing -> Nothing
    --   Just nc -> Just ( makeMove OutOf nc
    --                   , map (makeMove Into) ncs )

    makeMove :: Direction -> LNode Community -> LNode Community
    makeMove direction (cn, c) = (cn, moveNodeWithContext ctx direction c)

    -- k_i variable in formula (2)
    ki :: NodeWeightSum
    ki = nodeWeightSum ctx

    deltas :: [(LNode Community, DeltaQ)]
    deltas = map (\c -> (c, delta' c)) ncs

    delta' :: LNode Community -> DeltaQ
    delta' com = delta (llab com) ki kiin gw
      where
        -- k_i,in variable in formula (2)
        kiin :: NodeComWeightSum
        kiin = nodeComWeightSum (llab com) ctx

    -- TODO Compute \Delta Q (gain of moving node v into Community C) which consists of:
    -- - Community InWeightSum
    -- - Community TotWeightSum
    -- - sum of weights of links incident to node v (taken from 'p' and 's')
    -- - sum of weights of links from node v to nodes in C (taken from 'p' and 's')
    -- - sum of weights of all the links in the network
    -- So the Delta function takes as parameters:
    -- C, (edges from C in cgr), (edges from v in gr), (edges from v to C), (edges in gr)

-- COMMUNITY GRAPH FUNCTIONS

-- | 'Direction' when moving node 'Into'/'OutOf' community
data Direction = Into | OutOf

-- | Given 'Node' and 'Community' graph, find the 'LNode' of 'Community' which
-- contains the node
nodeCommunity :: Node -> CGr -> Maybe (LNode Community)
nodeCommunity n cgr = head (filter f $ labNodes cgr)
  where
    f :: (a, Community) -> Bool
    f (_, com) = n `elem` comNodes com

-- | Find 'LNode's of 'Community' graph neighbouring a given node
nodeNeighbours :: Node -> CGr -> [LNode Community]
nodeNeighbours n cgr =
  case nodeCommunity n cgr of
    Nothing        -> []
    Just (cn, _)  -> mapMaybe (lnode cgr) (neighbors cgr cn)

-- | Find 'Ajd CGrEdge's of 'Community' graph neighbouring a given node
-- nodeLNeighbours :: Node -> CGr -> Adj CGrEdge
-- nodeLNeighbours n cgr =
--   case nodeCommunity n cgr of
--     Nothing        -> []
--     Just (cn, _) -> lneighbors cgr cn

-- | Moves 'Node' in the 'Direction' of 'Community' and recomputes 'Community''s weights
moveNode :: forall a b. FGraph a b -> Node -> Direction -> Community -> Community
moveNode gr n direction c = moveNodeWithContext ctx direction c
  where
    ctx :: Context a (FEdge b)
    ctx = context gr n

-- | Same as 'moveNode' above but with only node context, not whole graph
moveNodeWithContext :: forall a b. Context a (FEdge b) -> Direction -> Community -> Community
moveNodeWithContext ctx@(_, n, _, _) direction com@(Community (ns, inwsum, totwsum)) =
  Community (newNs, InWeightSum newInWsum, TotWeightSum newTotWsum)
    where
      newNs = case direction of
        Into  -> sort (n:ns)
        OutOf -> DL.delete n ns

      (newInWsum, newTotWsum) =
        case direction of
          Into  -> (unInWeightSum inwsum + unNodeComWeightSum cws, unTotWeightSum totwsum - unNodeComWeightSum cws + unNodeNonComWeightSum nws)
          OutOf -> (unInWeightSum inwsum - unNodeComWeightSum cws, unTotWeightSum totwsum + unNodeComWeightSum cws - unNodeNonComWeightSum nws)

      -- sum of edge weights inside community
      cws :: NodeComWeightSum
      cws = nodeComWeightSum com ctx

      nws :: NodeNonComWeightSum
      nws = nodeNonComWeightSum com ctx


{-
-- | Moves 'Node' between two 'Community'ies, recomputing their weights
moveNode' :: FGraph a -> Node -> (Community, Community) -> (Community, Community)
moveNode' gr n cFromTo = moveNodeWithNeighbours' lnNeighbors n cFromTo
  where
    lnNeighbors :: Adj FEdge
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' but with direct neighbours list
moveNodeWithNeighbours' :: Adj FEdge -> Node -> (Community, Community) -> (Community, Community)
moveNodeWithNeighbours' lnNeighbors n ((Community (fromNs, fromInWSum, fromTotWSum)), (Community (toNs, toInWSum, toTotWSum))) =
  (newFrom, newTo)
  where
    newFrom = Community (DL.delete n fromNs, newFromInWSum, newFromTotWSum)
    newTo = Community (n:toNs, newToInWSum, newToTotWSum)

    -- Node is removed, so we reduce the internal weight sum
    newFromInWSum = fromInWSum - sumFromN
    newFromTotWSum = fromTotWSum + sumFromN - sumFromNonN

    -- Node is added, so we increase the internal weight sum
    newToInWSum = toInWSum + sumToN
    newToTotWSum = toTotWSum + sumToN - sumToNonN

    sumFromN :: InWeightSum
    sumFromN = sum $ map fst fromComNeighbors
    sumFromNonN :: TotWeightSum
    sumFromNonN = sum $ map fst fromNonComNeighbors

    sumToN :: InWeightSum
    sumToN = sum $ map fst toComNeighbors
    sumToNonN :: TotWeightSum
    sumToNonN = sum $ map fst toNonComNeighbors

    fromComNeighbors :: Adj FEdge
    fromComNeighbors = filter (\ln -> snd ln `elem` fromNs) lnNeighbors
    fromNonComNeighbors :: Adj FEdge
    fromNonComNeighbors = filter (\ln -> snd ln `notElem` fromNs) lnNeighbors
    toComNeighbors :: Adj FEdge
    toComNeighbors = filter (\ln -> snd ln `elem` toNs) lnNeighbors
    toNonComNeighbors :: Adj FEdge
    toNonComNeighbors = filter (\ln -> snd ln `notElem` toNs) lnNeighbors
-}
